import { __decorate, __metadata } from "tslib";
import { Event, Lazy, assertGet, joinPaths, typeGuard } from "@surface/core";
import Container from "@surface/dependency-injection";
import { painting } from "@surface/htmlx";
import { computed } from "@surface/observer";
import Router from "@surface/router";
import Metadata from "./metadata.js";
import RouteConfigurator from "./route-configurator.js";
const LEADING_SLASH_PATTERN = /^\//;
export default class WebRouter {
    baseUrl;
    connectedElements = [];
    container;
    history = [];
    interceptors;
    root;
    router = new Router();
    cache = [];
    index = 0;
    context;
    scope;
    routeChangeEvent = new Event();
    get route() {
        return this.context?.route
            ?? {
                meta: {},
                name: "",
                parameters: {},
                url: new URL(window.location.href),
            };
    }
    constructor(options) {
        this.root = new Lazy(() => assertGet(document.querySelector(options.root), `Cannot find root element using selector: ${options.root}`));
        this.baseUrl = options.baseUrl ?? /* c8 ignore next */ "";
        this.container = options.container ?? /* c8 ignore next */ new Container();
        this.interceptors = (options.interceptors ?? /* c8 ignore next */ []).map(x => typeof x == "function" ? this.container.inject(x) : x);
        this.scope = this.container.createScope();
        /* c8 ignore next */ // c8 can't cover iterable
        for (const definition of RouteConfigurator.configure(options.routes)) {
            this.router.map({
                constraints: definition.constraints,
                name: definition.name,
                pattern: joinPaths(this.baseUrl, definition.path),
                selector: routeData => [definition, routeData],
                transformers: definition.transformers,
            });
        }
        this.container.registerSingleton(WebRouter, this);
        window.addEventListener("popstate", () => void this.pushCurrentLocation());
    }
    connectToOutlet(parent, element, key, to, from, outletSelector = "router-outlet", reconnect = false) {
        const outlets = Metadata.from(parent).outlets;
        let outlet = outlets.get(key) ?? null;
        if (!outlet?.isConnected) {
            if (!parent.shadowRoot) {
                throw new Error("Routeable component requires an open shadowRoot");
            }
            outlet = parent.shadowRoot.querySelector(key == "default" ? `${outletSelector}:not([name])` : `${outletSelector}[name=${key}]`);
        }
        // istanbul ignore else
        if (outlet) {
            const router = this;
            outlet.remove = function () {
                this.parentNode?.removeChild(this);
                if (router.route == to) {
                    void painting().then(() => router.connectToOutlet(parent, element, key, to, from, outletSelector, true));
                }
            }.bind(outlet);
            outlets.set(key, outlet);
            const oldElement = outlet.firstElementChild;
            if (oldElement) {
                oldElement.onRouteLeave?.(to, from);
                oldElement.dispose?.();
                outlet.replaceChild(element, oldElement);
            }
            else {
                outlet.appendChild(element);
            }
            if (!reconnect) {
                element.onRouteEnter?.(to, from);
            }
            this.connectedElements.push(element);
        }
        else {
            throw new Error(`Cannot find outlet by provided selector "${outletSelector}${key == "default" ? "" : `[name=${key}]`}"`);
        }
    }
    async create(url, definition, routeData, fromHistory, replace) {
        const to = this.createRoute(url, definition, routeData);
        const from = this.context?.route;
        if (!await this.invokeMiddleware(to, from)) {
            const previous = this.context;
            const hasUpdate = definition == previous?.definition;
            const cache = this.cache.slice(0, definition.stack.length);
            this.context = { definition, route: to };
            let parent = this.root.value;
            let requireNewScope = true;
            for (let index = 0; index < definition.stack.length; index++) {
                const entry = definition.stack[index];
                const keys = new Set(entry.keys());
                this.disconnectFromOutlets(parent, keys, to, from);
                if (entry == previous?.definition?.stack[index]) {
                    const next = cache[index].default ?? cache[index][keys.values().next().value];
                    if (hasUpdate) {
                        next.onRouteUpdate?.(to, from);
                    }
                    else {
                        next.onRouteEnter?.(to, from);
                    }
                    requireNewScope = false;
                    parent = next;
                }
                else {
                    if (requireNewScope) {
                        this.scope.dispose();
                        this.scope = this.container.createScope();
                    }
                    let next;
                    cache[index] = {};
                    for (const [key, component] of entry) {
                        const constructor = await this.resolveComponent(component);
                        const element = this.scope.inject(constructor);
                        this.connectToOutlet(parent, element, key, to, from, definition.selector);
                        cache[index][key] = element;
                        if (!next || key == "default") {
                            next = element;
                        }
                    }
                    parent = next;
                }
            }
            this.cache = cache;
            if (!fromHistory) {
                if (replace) {
                    this.history[this.index] = [url, definition, routeData];
                }
                else {
                    this.history.splice(this.index + 1, Infinity);
                    this.history.push([url, definition, routeData]);
                    this.index = this.history.length - 1;
                }
            }
            this.routeChangeEvent.notify({ from, to });
        }
    }
    createRoute(url, definition, routeData) {
        return {
            meta: definition.meta,
            name: definition.name ?? "",
            parameters: routeData.parameters,
            url,
        };
    }
    disconnectElements() {
        this.connectedElements.forEach(x => (x.dispose?.(), x.remove()));
        this.context = undefined;
    }
    disconnectFromOutlets(parent, exclude, to, from) {
        const outlets = Metadata.from(parent).outlets;
        for (const outlet of outlets.values()) {
            if (!exclude.has(outlet.getAttribute("name") ?? "default")) {
                const instance = outlet.firstElementChild;
                if (typeGuard(instance, !!instance)) {
                    instance.onRouteLeave?.(to, from);
                    instance.dispose?.();
                    instance.remove();
                }
            }
        }
    }
    resolveModule(module) {
        return "default" in module ? module.default : module;
    }
    async resolveComponent(component) {
        if (typeof component == "function") {
            if (typeGuard(component, !component.prototype)) {
                return this.resolveComponent(component());
            }
        }
        else if (component instanceof Promise) {
            return this.resolveModule(await component);
        }
        return this.resolveModule(component);
    }
    async invokeMiddleware(to, from) {
        let intercepted = false;
        const next = async (location) => {
            await this.push(location);
            intercepted = true;
        };
        for (const interceptor of this.interceptors) {
            await interceptor.intercept(next, to, from);
            if (intercepted) {
                break;
            }
        }
        return intercepted;
    }
    resolvePath(path) {
        return new URL(path.replace(LEADING_SLASH_PATTERN, ""), `${joinPaths(window.location.origin, this.baseUrl)}/`);
    }
    async pushOrReplace(route, replace) {
        if (typeof route == "string") {
            const url = this.resolvePath(route);
            const match = this.router.match(url.pathname);
            if (match.matched) {
                await this.create(url, ...match.value, false, replace);
                window.history.pushState(null, "", this.route.url.href);
            }
            else {
                this.disconnectElements();
            }
        }
        else {
            const match = this.router.match(route.name, route.parameters ?? {});
            if (match.matched) {
                const [routeConfig, routeData] = match.value;
                const url = new URL(joinPaths(window.location.origin, routeData.path));
                url.hash = route.hash ?? "";
                for (const [key, value] of Object.entries(route.query ?? {})) {
                    if (Array.isArray(value)) {
                        value.forEach(x => url.searchParams.append(key, x));
                    }
                    else {
                        url.searchParams.set(key, value);
                    }
                }
                await this.create(url, routeConfig, routeData, false, replace);
                window.history.pushState(null, "", this.route.url.href);
            }
            else {
                this.disconnectElements();
            }
        }
    }
    async back() {
        await this.go(-1);
    }
    async forward() {
        await this.go(1);
    }
    async go(delta) {
        const index = Math.min(Math.max(this.index + delta, 0), this.history.length - 1);
        if (index != this.index) {
            const [url, routeItem, routeData] = this.history[index];
            await this.create(url, routeItem, routeData, true, false);
            window.history.replaceState(null, "", url.href);
            this.index = index;
        }
    }
    async push(route) {
        return this.pushOrReplace(route, false);
    }
    async replace(route) {
        return this.pushOrReplace(route, true);
    }
    async pushCurrentLocation() {
        await this.push(window.location.href);
    }
}
__decorate([
    computed("current"),
    __metadata("design:type", Object),
    __metadata("design:paramtypes", [])
], WebRouter.prototype, "route", null);
